/**
 *
 * Copyright 2005 LogicBlaze, Inc. http://www.logicblaze.com
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 **/

package org.logicblaze.ldap;

import org.logicblaze.ldap.bean.LDAPServerBean;

import javax.naming.*;
import javax.naming.directory.*;
import java.text.MessageFormat;
import java.util.Hashtable;

import com.liferay.portal.util.PrefsPropsUtil;
import com.liferay.portal.util.PropsUtil;


public class UserServiceUtil {

    public final static String KEY_LDAPSERVERINFO = "java:comp/env/ldap/apacheDS";

    public static void addUser(String userId, String password, String email,
                               String givenname, String surname) throws NamingException {

        if (userId == null || userId.length() < 1) {
            return;
        }

        LDAPServerBean server = null;
        DirContext context = null;

        try {
            //server = (LDAPServerBean) new InitialContext().lookup(KEY_LDAPSERVERINFO);
            server = getLDAPServerBean();
            context = getDirContext(server);

            String dn = "uid=" + userId + "," + server.getUserBase();

            EntityDirContext user = new EntityDirContext(dn, userId, password, email, givenname, surname);
            context.bind(dn, user);

        } catch (NamingException ne) {
            throw new NamingException("Error Adding User: " + ne.getMessage());

        } catch (Exception e) {
            throw new NamingException("Error in LDAP configuration: " + e.getMessage());

        } finally {
            closeDirContext(context);
        }
    }

    public static void updateUser(String userId, String email,
                                  String givenname, String surname) throws NamingException {

        if (userId == null || userId.length() < 1) {
            return;
        }

        LDAPServerBean server = null;
        DirContext context = null;

        try {
            //server = (LDAPServerBean) new InitialContext().lookup(KEY_LDAPSERVERINFO);
            server = getLDAPServerBean();
            context = getDirContext(server);

            Name name = getDistinguishedName(context, userId, server.getUserBase(), server.getUserSearchFilter());
            EntityDirContext user = new EntityDirContext(name.toString(), userId, email, givenname, surname);
            context.modifyAttributes(name, DirContext.REPLACE_ATTRIBUTE, user.getAttributes(""));

        } catch (NamingException ne) {
            throw new NamingException("Error Updating User: " + ne.getMessage());

        } catch (Exception e) {
            throw new NamingException("Error in LDAP configuration: " + e.getMessage());

        } finally {
            closeDirContext(context);
        }
    }

    public static void updatePassword(String userId, String password) throws NamingException {

        if (userId == null || userId.length() < 1) {
            return;
        }

        LDAPServerBean server = null;
        DirContext context = null;

        try {
            //server = (LDAPServerBean) new InitialContext().lookup(KEY_LDAPSERVERINFO);
            server = getLDAPServerBean();
            context = getDirContext(server);

            Name name = getDistinguishedName(context, userId, server.getUserBase(), server.getUserSearchFilter());
            EntityDirContext user = new EntityDirContext(name.toString(), userId, password);
            context.modifyAttributes(name, DirContext.REPLACE_ATTRIBUTE, user.getAttributes(""));

        } catch (NamingException ne) {
            throw new NamingException("Error Updating Password: " + ne.getMessage());

        } catch (Exception e) {
            throw new NamingException("Error in LDAP configuration: " + e.getMessage());

        } finally {
            closeDirContext(context);

        }
    }

    public static void deleteUser(String userId) throws NamingException {

        if (userId == null || userId.length() < 1) {
            return;
        }

        LDAPServerBean server = null;
        DirContext context = null;

        try {
            //server = (LDAPServerBean) new InitialContext().lookup(KEY_LDAPSERVERINFO);
            server = getLDAPServerBean();
            context = getDirContext(server);

            Name name = getDistinguishedName(context, userId, server.getUserBase(),
                    server.getUserSearchFilter());
            context.destroySubcontext(name);

        } catch (NamingException ne) {
            throw new NamingException("Error Deleting User: " + ne.getMessage());

        } catch (Exception e) {
            throw new NamingException("Error in LDAP configuration: " + e.getMessage());

        } finally {
            closeDirContext(context);
        }
    }

    public static void addRole(String roleId) throws NamingException {

        if (roleId == null || roleId.length() < 1) {
            return;
        }

        LDAPServerBean server = null;
        DirContext context = null;

        try {
            //server = (LDAPServerBean) new InitialContext().lookup(KEY_LDAPSERVERINFO);
            server = getLDAPServerBean();
            context = UserServiceUtil.getDirContext(server);

            String dn = "cn=" + roleId + "," + server.getRoleBase();

            EntityDirContext role = new EntityDirContext(dn, roleId);

            context.bind(dn, role);

        } catch (NamingException ne) {
            throw new NamingException("Error Adding Role: " + ne.getMessage());

        } catch (Exception e) {
            throw new NamingException("Error in LDAP configuration: " + e.getMessage());

        } finally {
            UserServiceUtil.closeDirContext(context);
        }
    }

    public static void addRoleUsers(String roleId, String[] userIds) throws NamingException {

        if (roleId == null || roleId.length() < 1 ||
                userIds == null || userIds.length < 1) {

            return;
        }

        LDAPServerBean server = null;
        DirContext context = null;

        try {
            //server = (LDAPServerBean) new InitialContext().lookup(KEY_LDAPSERVERINFO);
            server = getLDAPServerBean();
            context = UserServiceUtil.getDirContext(server);

            Name role = UserServiceUtil.getDistinguishedName(context, roleId, server.getRoleBase(),
                    server.getRoleSearchFilter());

            ModificationItem[] mods = new ModificationItem[1];
            Attribute members = new BasicAttribute("uniquemember");

            int counter = 0;
            for (int i = 0; i < userIds.length; i++) {

                boolean isMember = isMember(context, server, roleId, userIds[i]);

                if (!isMember) {
                    members.add("uid=" + userIds[i] + "," + server.getUserBase());
                    counter++;
                }
            }

            // Execute only when there is members to be added.
            if (counter > 0) {
                mods[0] = new ModificationItem(DirContext.ADD_ATTRIBUTE, members);
                context.modifyAttributes(role, mods);
            }


        } catch (NamingException ne) {
            throw new NamingException("Error Adding Users to Role: " + ne.getMessage());

        } catch (Exception e) {
            throw new NamingException("Error in LDAP configuration: " + e.getMessage());

        } finally {
            UserServiceUtil.closeDirContext(context);
        }
    }

    public static void unsetRoleUsers(String roleId, String[] userIds) throws NamingException {

        if (roleId == null || roleId.length() < 1 ||
                userIds == null || userIds.length < 1) {

            return;
        }

        LDAPServerBean server = null;
        DirContext context = null;

        try {
            //server = (LDAPServerBean) new InitialContext().lookup(KEY_LDAPSERVERINFO);
            server = getLDAPServerBean();
            context = UserServiceUtil.getDirContext(server);

            Name role = UserServiceUtil.getDistinguishedName(context, roleId, server.getRoleBase(),
                    server.getRoleSearchFilter());

            ModificationItem[] mods = new ModificationItem[1];
            Attribute members = new BasicAttribute("uniquemember");

            boolean hasMultipleMembers = hasMultipleMembers(context, server, roleId);

            // Execute if the role has only 1 member.
            if (!hasMultipleMembers) {
                for (int i = 0; i < userIds.length; i++) {
                    if (isMember(context, server, roleId, userIds[i])) {
                        members.add("");
                        mods[0] = new ModificationItem(DirContext.REPLACE_ATTRIBUTE, members);
                        context.modifyAttributes(role, mods);
                        return;
                    }
                }
            }

            for (int i = 0; i < userIds.length; i++) {

                boolean isMember = isMember(context, server, roleId, userIds[i]);

                if (isMember) {
                    members.add("uid=" + userIds[i] + "," + server.getUserBase());
                }

            }

            // Execute update only when there is members to be updated.
            if (members.size() > 0) {
                mods[0] = new ModificationItem(DirContext.REMOVE_ATTRIBUTE, members);
                context.modifyAttributes(role, mods);
            }

        } catch (NamingException ne) {
            throw new NamingException("Error Removing Users from Role: " + ne.getMessage());

        } catch (Exception e) {
            throw new NamingException("Error in LDAP configuration: " + e.getMessage());

        } finally {
            UserServiceUtil.closeDirContext(context);

        }
    }

    // --------------------------------------------
    //  HELPER METHODS
    // --------------------------------------------

    protected static Name getDistinguishedName(DirContext context, String base,
                                               String filter, boolean searchSubtree) throws NamingException {

        SearchControls constraints = new SearchControls();

        if (searchSubtree) {
            constraints.setSearchScope(SearchControls.SUBTREE_SCOPE);
        } else {
            constraints.setSearchScope(SearchControls.ONELEVEL_SCOPE);
        }

        NamingEnumeration results = context.search(base, filter, constraints);

        if (results == null || !results.hasMore()) {
            return null;
        }

        SearchResult result = (SearchResult) results.next();

        Attributes attrs = result.getAttributes();

        if (attrs == null)
            return null;

        NameParser parser = context.getNameParser("");
        Name contextName = parser.parse(context.getNameInNamespace());

        Name baseName = parser.parse(base);
        Name name = contextName.addAll(baseName);

        Name entryName = parser.parse(result.getName());
        name = name.addAll(entryName);

        return name;
    }

    protected static Name getDistinguishedName(DirContext context, String id,
                                               String entityBase, String searchFilter)
            throws NamingException {

        MessageFormat searchMatchingFormat = new MessageFormat(searchFilter);
        String filter = searchMatchingFormat.format(new String[]{id});
        return getDistinguishedName(context, entityBase, filter, true);
    }

    public static boolean isMember(DirContext context, LDAPServerBean server, String roleId, String userId) throws NamingException {

        try {
            Attribute cn = new BasicAttribute("cn");
            cn.add(roleId);

            Attribute member = new BasicAttribute("uniquemember");

            member.add("uid=" + userId + "," + server.getUserBase());

            Attributes myAttrs = new BasicAttributes(true);
            myAttrs.put(cn);
            myAttrs.put(member);

            NamingEnumeration result = context.search(server.getRoleBase(), myAttrs);

            while (result.hasMore()) {
                SearchResult sr = (SearchResult) result.next();

                if (("cn=" + roleId).equalsIgnoreCase(sr.getName())) {
                    return true;
                }
            }

        } catch (NamingException ne) {
            throw new NamingException();

        }

        return false;
    }

    public static boolean hasMultipleMembers(DirContext context, LDAPServerBean server, String roleId) throws NamingException {

        try {
            Attribute cn = new BasicAttribute("cn");
            cn.add(roleId);

            Attribute member = new BasicAttribute("uniqueMember");

            Attributes myAttrs = new BasicAttributes(true);
            myAttrs.put(cn);
            myAttrs.put(member);

            NamingEnumeration result = context.search(server.getRoleBase(), myAttrs);

            while (result.hasMore()) {
                SearchResult sr = (SearchResult) result.next();
                Attributes attrs = sr.getAttributes();
                Attribute attr = attrs.get("uniquemember");

                NamingEnumeration e1 = attr.getAll();

                int ctr = 0;
                while (e1.hasMore()) {
                    e1.nextElement();

                    if (ctr > 1) {
                        return true;
                    }
                    ctr++;
                }

                return false;
            }

        } catch (NamingException ne) {
            throw new NamingException();

        }


        return false;
    }

    protected static DirContext getDirContext(LDAPServerBean server) throws NamingException {

        Hashtable env = new Hashtable();
        env.put(Context.INITIAL_CONTEXT_FACTORY, server.getInitialContextFactory());
        env.put(Context.SECURITY_PRINCIPAL, server.getSecurityPrincipal());
        env.put(Context.SECURITY_CREDENTIALS, server.getSecurityCredentials());
        env.put(Context.SECURITY_PROTOCOL, server.getSecurityProtocol());
        env.put(Context.PROVIDER_URL, server.getProviderUrl());
        env.put(Context.SECURITY_AUTHENTICATION, server.getSecurityAuthentication());

        return new InitialDirContext(env);
    }

    protected static void closeDirContext(DirContext context) {

        try {
            context.close();
            context = null;

        } catch (NamingException e) {
            e.printStackTrace();

        }
    }


    public static LDAPServerBean getLDAPServerBean() throws Exception {


        LDAPServerBean serverInfo = new LDAPServerBean();
        serverInfo.setInitialContextFactory(PrefsPropsUtil.getString(com.liferay.portal.util.PropsUtil.AUTH_IMPL_LDAP_FACTORY_INITIAL));
        serverInfo.setProviderUrl(PrefsPropsUtil.getString(com.liferay.portal.util.PropsUtil.AUTH_IMPL_LDAP_PROVIDER_URL));
        serverInfo.setRoleBase(PrefsPropsUtil.getString(com.liferay.portal.util.PropsUtil.AUTH_IMPL_LDAP_ROLEBASE));
        serverInfo.setRoleSearchFilter(PrefsPropsUtil.getString(com.liferay.portal.util.PropsUtil.AUTH_IMPL_LDAP_ROLESEARCH_FILTER));
        serverInfo.setSecurityAuthentication(PrefsPropsUtil.getString(com.liferay.portal.util.PropsUtil.AUTH_IMPL_LDAP_SECURITY_AUTHENTICATION));
        serverInfo.setSecurityCredentials(PrefsPropsUtil.getString(com.liferay.portal.util.PropsUtil.AUTH_IMPL_LDAP_SECURITY_CREDENTIALS));
        serverInfo.setSecurityPrincipal(PrefsPropsUtil.getString(com.liferay.portal.util.PropsUtil.AUTH_IMPL_LDAP_SECURITY_PRINCIPAL));
        serverInfo.setSecurityProtocol(PrefsPropsUtil.getString(com.liferay.portal.util.PropsUtil.AUTH_IMPL_LDAP_SECURITY_PROTOCOL));
        serverInfo.setUserBase(PrefsPropsUtil.getString(com.liferay.portal.util.PropsUtil.AUTH_IMPL_LDAP_USERBASE));
        serverInfo.setUserSearchFilter(PrefsPropsUtil.getString(com.liferay.portal.util.PropsUtil.AUTH_IMPL_LDAP_USERSEARCH_FILTER));

        return serverInfo;
    }

}
